unit CoreExports;

//The di driver uses "array" to refer to user-allocated RAM, and "buffer" to refer to driver-allocated RAM.
//0 is analog input buffer, 1 is analog output buffer.

//<http://www.dataq.com/resources/installations/web-help/sdk/>

interface

uses
	Windows;

var
	{
	bLogging: Boolean = True;
	bDateTimeLog: Boolean = False;
	bLogCalls: Boolean = True;
	bLogStats: Boolean = True;
	bLogConfigFile: Boolean = False;
	FinalDataLog: (fdlNone, fdlRaw, fdlRawDecimal, fdlParsed) = fdlNone;
	BufferStatusLog: (bslNone, bslOnChange, bslAlways) = bslNone;
	bLogCallbackMinMax: Boolean = False;
	bLogCallbackDecimalRaw: Boolean = False;
	}

	bLogging: Boolean = False;
	bDateTimeLog: Boolean = False;
	bLogCalls: Boolean = False;
	bLogStats: Boolean = False;
	bLogConfigFile: Boolean = False;
	FinalDataLog: (fdlNone, fdlRaw, fdlRawDecimal, fdlParsed) = fdlNone;
	BufferStatusLog: (bslNone, bslOnChange, bslAlways) = bslNone;
	bLogCallbackMinMax: Boolean = False;
	bLogCallbackDecimalRaw: Boolean = False;

	bLoadConfigOnListLen: Boolean = False;
	DebugDelay: Integer = -1; //100;
	HammerExtra: Byte = 0; //7;

const
	BaseDACCalFactor = 0.9803;
	ADCCalScale = -10;
	DACCalScale = +10;

type
	//Dataq API structs.
	TDIAnInData = packed record
		Chan: Word; //Channel number ranging from 0 to 15
		Diff: Word; //0 = single-ended; 1 = differential
		Gain: Word; //Gain code ranging from 0 (lowest gain) to 3 (highest gain)
		Unipolar: Word; //0 = bipolar; 1 = unipolar
	end;

	TDIInfoData16 = packed record
		Port: Word; //Device port address
		buf_in_chn: Word; //Device input channel
		buf_out_chn: Word; //Device output channel
		sft_lvl: Word; //Software interrupt level
		hrd_lvl: Word; //Hardware interrupt level
		Pad0: Word;
		buf_in_ptr: LongWord; //Input buffer pointer
		buf_in_size: Word; //Input buffer size
		Pad1: Word;
		buf_out_ptr: LongWord; //Output buffer pointer
		buf_out_size: Word; //Output buffer size
		Pad2: Word;
		tsr_version: array[0..19] of AnsiChar; //TSR version and model number
		dsp_version: array[0..19] of AnsiChar; //DSP program version
		sdk_version: array[0..19] of AnsiChar; //SDK library version
		serial_no: LongWord; //Hardware serial number (unsigned long)
		last_cal: LongWord; //Date and time of last calibration (unsigned long)
		board_id: array[0..9] of AnsiChar; //Hardware model number
		pgh_pgl: Byte; //1=PGH; 0=PGL
		hrdwr_rev: Byte; //Hardware revision level
		pal0_rev: Byte; //PAL0 revision level
		pal1_rev: Byte; //PAL1 revision level
	end;

	TDIInfoData32 = record //Use default 32-bit padding.
		Port: Word; //Device port address
		buf_in_chn: Word; //Device input channel
		buf_out_chn: Word; //Device output channel
		sft_lvl: Word; //Software interrupt level
		hrd_lvl: Word; //Hardware interrupt level
		buf_in_ptr: LongWord; //Input buffer pointer
		buf_in_size: Word; //Input buffer size
		buf_out_ptr: LongWord; //Output buffer pointer
		buf_out_size: Word; //Output buffer size
		tsr_version: array[0..19] of AnsiChar; //TSR version and model number
		dsp_version: array[0..19] of AnsiChar; //DSP program version
		sdk_version: array[0..19] of AnsiChar; //SDK library version
		serial_no: LongWord; //Hardware serial number (unsigned long)
		last_cal: LongWord; //Date and time of last calibration (unsigned long)
		board_id: array[0..9] of AnsiChar; //Hardware model number
		pgh_pgl: Byte; //1=PGH; 0=PGL
		hrdwr_rev: Byte; //Hardware revision level
		pal0_rev: Byte; //PAL0 revision level
		pal1_rev: Byte; //PAL1 revision level
	end;

	TDIInlistData = packed record
		Chan: Word; //Input channel (0-15)
		Diff: Word; //0=single-ended; 1=differential
		//Set Chan=8 and Diff=1 for digital input.  Otherwise the list defined by the structure accepts analog input.
		Gain: Word; //Gain code ranging from 0 (lowest gain) to 3 (highest gain)
		Unipolar: Word; //1=unipolar; 0= bipolar
		dig_out_enable: Word; //1= digital out enabled; 0=digital out disabled
		dig_out: Word; //D0-D4=data
		Ave: Word; //Sample averaging, 0=off; 1=on
		counter: Word; //Sample rate counter (0-32,767)
	end;
	TPDIInlistData = ^TDIInlistData;

	TDIOutlistData = packed record
		Unipolar: Word; //1=unipolar; 0= bipolar
		digital: Word; //Digital output=1; analog output=0
		dig_out_enable: Word; //1= enable digital out; 0=disable digital out
		dig_data: Word; //Digital data to output to D8-D15
		counter: Word; //Output rate counter (0-32767)
	end;
	TPDIOutlistData = ^TDIOutlistData;

function AIO_AddCal(CalVB, CalIB: Double): LongWord; stdcall;

function di_open(USBID: LongWord): Word; stdcall;
function di_close(): Word; stdcall;

function di_info(var A1: TDIInfoData16): Word; stdcall;
function di_info32(var A1: TDIInfoData32): Word; stdcall;

function di_anin(const AninData: TDIAnInData): SmallInt; stdcall;
function di_anout(DACData: SmallInt; Channel: Word): Word; stdcall;

function di_digin(): Word; stdcall;
function di_digout(DIOData: Word): Word; stdcall;

function di_buffer_alloc(Dir, BufferWords: Word): PWord; stdcall;
function di_buffer_free(Dir: Word): Word; stdcall;

function di_list_length(InLength, OutLength: Word): Word; stdcall;
function di_inlist(pEntry: TPDIInlistData): Word; stdcall;
function di_outlist(pEntry: TPDIOutlistData): Word; stdcall;
function di_burst_rate(Divisor: Word): Word; stdcall;
function di_start_scan(): Word; stdcall;
function di_stop_scan(): Word; stdcall;

function di_buffer_status(Dir: Word): LongWord; stdcall;
function di_copy_buffer(BufIndex: Word; pDest: Pointer; Count: Word): Word; stdcall;
function di_copy_array(BufIndex: Word; pSource: PWord; Count: Word): Word; stdcall;





implementation

uses
	Math, FileCtrl, SysUtils, Classes,
	DFileStreams, AIOUSB,
	AbortThreads, TheseUtils, LoggingUtils;

const
	DI_NO_ERR = 0;
	DI_DRIVER_ERR = 1;
	DI_COMM_ERR = 2;
	DI_RESET_ERR = 3;
	DI_CLOSE_ERR = 4;
	DI_INFO_ERR = 5;
	DI_DIGOUT_ERR = 6;
	DI_ANOUT_ERR = 7;
	DI_MODE_ERR = 8;
	DI_START_SCN_ERR = 9;
	DI_STOP_SCN_ERR = 10;
	DI_INLIST_ERR = 11;
	DI_OUTLIST_ERR = 12;
	DI_BURST_ERR = 13;
	DI_LENGTH_ERR = 14;
	DI_EVENT_COUNT_ERR = 15;
	DI_OPENED_ERR = 19;
	DI_ERR_CODE = 20;
	DI_ALREADY_OPEN_ERR = 21;

type
	TRingBufferData = record
		Buf: array of Word;
		BufLen, NextIndex: Integer;
		hIndexMutex: THandle;
	end;

	TSaturationType = (stLower, stUpper);
	TSaturationTypes = set of TSaturationType;

	TDeviceData = record
		ShadowConfig: array[0..$14] of Byte;
		SN: Int64;

		NumChannels, TheirSamplesPerChannel, SamplesPerChannel, DiscardOversamples: Integer;
		ScanHz, BurstHz: Double;
		DACChannel: Word;
		InListLen, OutListLen: Integer;
		AIBuf, AOBuf: TRingBufferData;
		bScanning, bSuspendedByMisuse, bRejectingData, bWriteHead, bGotCallback: Boolean;
		LastCallbackTick, TicksPer3Callbacks: Int64;
		BeginningOfTheEnd: Integer;
		SmoothingMode: (smAverage, smMedian, smAverageTheMiddle, smAsymmetricStandardDeviations, smInertia);
		SmoothingMiddle: Double;
		iSmoothingCull, SmoothingASDReps: Integer;
		SmoothingInertia: Integer;

		ThisAIOversample, ThisAICh: Integer;
		IncompleteChannel: Integer;
		IncompleteScan: array of Word;
		Hist: array of Word; //Histogram for radix sort.
		AvgFactor: Double;
		EID: LongWord;
		InertialChData: array[$0..$F] of Integer;

		bSaturatedPulse: Boolean;
		OuterDataOriginal: array[$0..$F] of Integer;

		hDACLastMutex: THandle;
		DACLastOut: array[0..1] of Integer;
		bAOStopped: Boolean;
		DACBufferStuffed, LastDACReport: Integer;

		//For logging, and for median-y smoothing modes.
		nData, xData: Word;
		ChSaturation: array of TSaturationTypes;
		AccScanStr: AnsiString;
		RawScanStr: AnsiString;
		LastBufferStatus: Integer;
	end;
	TPDeviceData = ^TDeviceData;

	TCalPair = record
		M, B: Double;
	end;

var
	hCardMutex: THandle;
	Dev: TDeviceData;
	DeviceIndex: LongWord = diNone;
	DACDeviceIndex: LongWord = diNone;
	DACBase: Integer;
	TicksPerSec: Int64;
	ConfigFileName: AnsiString;
	ConfigADCal: array[Boolean] of TCalPair;
	ConfigDACCal: TCalPair;
	RangeCal: array[0..7] of TCalPair;
	OutChCal: array[$0..$F] of TCalPair;
	ChGainFactor: array[$0..$F] of Integer;
	SpeedCode: Integer;

const
	DirToStr: array[0..1] of AnsiString = ('A/D', 'D/A');
	CardMutexNom = 'AIO_8140_CARD_MUTEX';
	CountsPerVolt = $10000/20;





function BalanceAsymmetricStandardDeviations(PrevApprox: Double; MethodReps: Integer = 1): Double; forward;
procedure ADContCallback(pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl; forward;





function DIGainCodeToRangeCode(DI_Gain: Word): Byte;
begin
	case DI_Gain of
		0: Result := 1;
		1: Result := 3;
		2: Result := 5; //Use 2 for 2.5.
		3: Result := 7; //Use 1 for 1.25.
		else Result := 1; //Fall back to 10.
	end;
end;

procedure _StartOutputFreq(DeviceIndex, BlockIndex: LongWord; pHz: PDouble); cdecl;
const
	RootClock = 10000000;

var
	L, DivisorA, DivisorB: LongWord;
	DivisorAB: Extended;
	Err, MinErr, Hz: Double;

begin
	if pHz = nil then begin
		SetLastError(ERROR_INVALID_PARAMETER);
		Exit;
	end;

	Hz := pHz^;
	if Hz <= 0 then begin
		CTR_8254Mode(DeviceIndex, BlockIndex, 1, 2);
		Sleep(1);
		CTR_8254Mode(DeviceIndex, BlockIndex, 2, 2);
	end
	else begin
		try
			if Hz * 4 >= RootClock then begin
				DivisorA := 2;
				DivisorB := 2;
			end
			else begin
				DivisorAB := RootClock / Hz;

				L := Round(SqRt(DivisorAB));
				DivisorA := Round(DivisorAB / L);
				DivisorB := L;
				MinErr := Abs(Hz - (RootClock / (DivisorA * L)));
				for L := L downto 2 do begin
					DivisorA := Round(DivisorAB / L);
					if DivisorA > $FFFF then Break; //Limited to 16 bits, so this and all further L are invalid.
					Err := Abs(Hz - (RootClock / (DivisorA * L)));
					if Err = 0 then begin
						DivisorB := L;
						Break;
					end;
					if Err < MinErr then begin
						DivisorB := L;
						MinErr := Err;
					end;
				end;
				DivisorA := Round(DivisorAB / DivisorB);
			end;
		except
			SetLastError(ERROR_INVALID_DATA);
			Exit;
		end;

		pHz^ := RootClock / (DivisorA * DivisorB);

		CTR_8254ModeLoad(DeviceIndex, BlockIndex, 1, 2, DivisorA);
		Sleep(1);
		CTR_8254ModeLoad(DeviceIndex, BlockIndex, 2, 2, DivisorB);
	end;
end;

procedure FreqOut(Hz: Double);
var
	I: Integer;

begin
	if bLogging then LogCall('FreqOut(' + IntToHex(Round(Hz), 1) + ')');

	if DebugDelay >= 0 then Sleep(DebugDelay);
	for I := 0 to HammerExtra do _StartOutputFreq(DeviceIndex, 0, @Hz);
end;

function DACOut(Channel: Integer; Data: Double; bThreadLog: Boolean = False): LongWord;
var
	I: Integer;
	bWriteNow: Boolean;
	SignSamp: SmallInt;
	DACData: Word;
	LogStr: AnsiString;

begin
	Result := ERROR_SUCCESS;

	with ConfigDACCal do Data := M * (Data + B); //Note: M*(X+B) instead of (M*X)+B so it's easier to type in numbers.
	if Data >= High(SignSamp) then
		SignSamp := High(SignSamp)
	else if Data <= Low(SignSamp) then
		SignSamp := Low(SignSamp)
	else
		SignSamp := Round(Data)
	;
	DACData := Word(SignSamp) xor $8000;

	WaitForSingleObject(Dev.hDACLastMutex, INFINITE);
		bWriteNow := DACData <> Dev.DACLastOut[Channel];
		if bWriteNow then Dev.DACLastOut[Channel] := DACData;
	ReleaseMutex(Dev.hDACLastMutex);

	if not bWriteNow then Exit;

	if DebugDelay >= 0 then Sleep(DebugDelay);
	for I := 0 to HammerExtra do Result := DACDirect(DACDeviceIndex, DACBase+Channel, DACData);
	if bLogging then begin
		LogStr := 'DACDirect(' + IntToHex(DACDeviceIndex, 1) + ', ' + IntToHex(DACBase+Channel, 1) + ', ' + IntToHex(DACData, 4) + ')';
		if bThreadLog then
			ThreadLogCall(LogStr)
		else
			LogCall(LogStr)
		;
	end;
end;

function SetConfig: LongWord;
const
	RangeCodeToGainFactor: array[$0..$7] of Integer = (2, 1, 4, 2, 10, 5, 20, 10);

var
	ConfigSize: LongWord;
	I: Integer;
	LogStr: AnsiString;

begin
	ConfigSize := Length(Dev.ShadowConfig);
	Result := ADC_SetConfig(DeviceIndex, @Dev.ShadowConfig[0], ConfigSize);

	for I := 0 to 15 do
		ChGainFactor[I] := RangeCodeToGainFactor[Dev.ShadowConfig[I] and $7];
	;

	if bLogging then begin
		LogStr := 'ADC_SetConfig(' + IntToHex(DeviceIndex, 1) + ', [' + IntToHex(Dev.ShadowConfig[0], 2);
		for I := 1 to High(Dev.ShadowConfig) do
			LogStr := LogStr + (', ' + IntToHex(Dev.ShadowConfig[I], 2))
		;
		LogCall(LogStr + '], ' + IntToHex(ConfigSize, 1) + ') = ' + IntToStr(Result));
	end;
end;

function ElectSerialNumber(DeviceIndex: LongWord; var SN: Int64): LongWord;
var
	ErrTol, NCap, I, N: Integer;
	Status: LongWord;
	BufSN: Int64;
	NominatedSN: array of Int64;

begin
	SN := 0;
	ErrTol := 8;
	Result := ERROR_SUCCESS;
	N := 0;
	NCap := 4;
	SetLength(NominatedSN, NCap);
	repeat
		Status := GetDeviceSerialNumber(DeviceIndex, BufSN);
		if Status <> ERROR_SUCCESS then begin
			if Result = ERROR_SUCCESS then Result := Status;
			Dec(ErrTol);
			if ErrTol <= 0 then Exit;
			Continue;
		end;

		for I := 0 to N-1 do
			if NominatedSN[I] = BufSN then begin
				SN := BufSN;
				Result := ERROR_SUCCESS;
				Exit;
			end
		;

		if NCap <= N then begin
			NCap := 2 * NCap;
			SetLength(NominatedSN, NCap);
		end;
		NominatedSN[N] := BufSN;
		Inc(N);
	until False;
end;

procedure SetSpeeds(bAlwaysSetConfig: Boolean = False);
const
	MaxSamplesPerChannel = 256;

begin
	with Dev do begin
		ScanHz := BurstHz / (TheirSamplesPerChannel * Max(InListLen, OutListLen));
		SamplesPerChannel := 128; //Min(MaxSamplesPerChannel, Floor(500000 / (ScanHz * NumChannels))); //How much oversample we can do.
		DiscardOversamples := Min(40, SamplesPerChannel div 2);
		AvgFactor := 1 / (SamplesPerChannel - DiscardOversamples);

		if bAlwaysSetConfig or (ShadowConfig[$13] <> SamplesPerChannel - 1) then begin
			ShadowConfig[$13] := SamplesPerChannel - 1;
			SetConfig;
		end;

		if bLogging and bLogStats then begin
			if DiscardOversamples <> 0 then
				LogCall('Oversample x' + IntToHex(SamplesPerChannel, 1) + 'k' + IntToHex(SamplesPerChannel - DiscardOversamples, 1))
			else
				LogCall('Oversample x' + IntToHex(SamplesPerChannel, 1))
			;
		end;
	end;
end;

procedure ParseVStrAsB(var Result: Double; Scale: Double; VStr: AnsiString);
var
	L: Integer;

begin
	L := Length(VStr);
	if Copy(VStr, L-1, 2) = 'mV' then
		Result := CountsPerVolt/1000 * Scale * StrToFloat(Trim(Copy(VStr, 1, L-2)))
	else if Copy(VStr, L, 1) = 'V' then
		Result := CountsPerVolt * Scale * StrToFloat(Trim(Copy(VStr, 1, L-1)))
	else if Copy(VStr, L-1, 2) = 'ct' then
		Result := Scale * StrToFloat(Trim(Copy(VStr, 1, L-2)))
	else //Assume millivolts.
		Result := CountsPerVolt/1000 * Scale * StrToFloat(Trim(VStr))
	;
end;

function StrToProportionFloat(S: AnsiString; MinFrac: Double = 0; MaxFrac: Double = 1): Double;
var
	I: Integer;
	bPercent: Boolean;

begin
	I := Pos('%', S);
	bPercent := I <> 0;
	if bPercent then S := Copy(S, 1, I-1);
	Result := StrToFloat(S);
	if bPercent then Result := 1/100 * Result;
	if Result > MaxFrac then Result := MaxFrac;
	if Result < MinFrac then Result := MinFrac;
end;

function BoolToStr(B: Boolean): AnsiString;
begin
	if B then
		Result := 'True'
	else
		Result := 'False'
	;
end;

procedure DoLoadConfigFile;
var
	I, iLine: Integer;
	Content, LValue, RValue, ThisSection: AnsiString;
	Contents: TStringList;

begin
	if bLogging and bLogConfigFile then LogCall('Loading config file "' + ConfigFileName + '"');
	ConfigADCal[False].M := 1;
	ConfigADCal[False].B := 0;
	ConfigADCal[True].M := 1;
	ConfigADCal[True].B := 0;
	ConfigDACCal.M := BaseDACCalFactor;
	ConfigDACCal.B := 0;
	SpeedCode := -1;
	Dev.SmoothingMode := smAsymmetricStandardDeviations;
	Dev.SmoothingMiddle := 3/5;
	Dev.SmoothingASDReps := 2;
	Dev.SmoothingInertia := 64;
	if FileExists(ConfigFileName) then begin
		Contents := TStringList.Create;
		try
			Contents.LoadFromFile(ConfigFileName);
			ThisSection := '';
			for iLine := 0 to Contents.Count - 1 do begin
				Content := Trim(Contents[iLine]);
				if Content = '' then Continue;
				if Content[1] = ';' then Continue;

				if (Content[1] = '[') and (Content[Length(Content)] = ']') then begin
					//Sections.
					ThisSection := Content;
					Continue;
				end;

				I := Pos('=', Content);
				if I <> 0 then begin
					//Assign commands.
					LValue := UpperCase(TrimRight(Copy(Content, 1, I-1)));
					RValue := TrimLeft(Copy(Content, I+1, MAXINT));

					if ThisSection = '[ADC]' then begin
						if LValue = 'CALM' then try
							ConfigADCal[False].M := StrToFloat(RValue);
							ConfigADCal[True].M := ConfigADCal[False].M;
							if bLogging and bLogConfigFile then LogCall('[ADC] CalM = ' + FloatToStr(ConfigADCal[False].M) + '.');
						except
						end
						else if LValue = 'CALVM' then try
							ConfigADCal[False].M := StrToFloat(RValue);
							if bLogging and bLogConfigFile then LogCall('[ADC] CalVM = ' + FloatToStr(ConfigADCal[False].M) + '.');
						except
						end
						else if LValue = 'CALIM' then try
							ConfigADCal[True].M := StrToFloat(RValue);
							if bLogging and bLogConfigFile then LogCall('[ADC] CalIM = ' + FloatToStr(ConfigADCal[True].M) + '.');
						except
						end
						else if LValue = 'CALB' then try
							ParseVStrAsB(ConfigADCal[False].B, ADCCalScale, RValue);
							ConfigADCal[True].B := ConfigADCal[False].B;
							if bLogging and bLogConfigFile then LogCall('[ADC] CalB = ' + FloatToStr(ConfigADCal[False].B) + '.');
						except
						end
						else if LValue = 'CALVB' then try
							ParseVStrAsB(ConfigADCal[False].B, ADCCalScale, RValue);
							if bLogging and bLogConfigFile then LogCall('[ADC] CalVB = ' + FloatToStr(ConfigADCal[False].B) + '.');
						except
						end
						else if LValue = 'CALIB' then try
							ParseVStrAsB(ConfigADCal[True].B, ADCCalScale, RValue);
							if bLogging and bLogConfigFile then LogCall('[ADC] CalIB = ' + FloatToStr(ConfigADCal[True].B) + '.');
						except
						end
						else if LValue = 'SPEEDCODE' then try
							SpeedCode := StrToInt('$' + RValue);
							if bLogging and bLogConfigFile then LogCall('[ADC] SpeedCode = ' + IntToHex(SpeedCode, 2) + '.');
						except
						end
						else if LValue = 'SMOOTHING' then try
							RValue := UpperCase(Trim(RValue));
							if RValue = 'AVERAGE' then begin
								Dev.SmoothingMode := smAverage;
								if bLogging and bLogConfigFile then LogCall('[ADC] smoothing mode is average.');
							end
							else if RValue = 'MEDIAN' then begin
								Dev.SmoothingMode := smMedian;
								if bLogging and bLogConfigFile then LogCall('[ADC] smoothing mode is median.');
							end
							else if RValue = 'AVERAGETHEMIDDLE' then begin
								Dev.SmoothingMode := smAverageTheMiddle;
								if bLogging and bLogConfigFile then LogCall('[ADC] smoothing mode is average-the-middle.');
							end
							else if RValue = 'ASD' then begin
								Dev.SmoothingMode := smAsymmetricStandardDeviations;
								if bLogging and bLogConfigFile then LogCall('[ADC] smoothing mode is asymmetric standard deviations.');
							end
							else if RValue = 'INERTIA' then begin
								Dev.SmoothingMode := smInertia;
								if bLogging and bLogConfigFile then LogCall('[ADC] smoothing mode is inertia.');
							end
							else begin
								if bLogging and bLogConfigFile then LogCall('Unrecognized [ADC] smoothing mode R-value "' + RValue + '".');
							end;
						except
						end
						else if LValue = 'SMOOTHINGMIDDLE' then try
							Dev.SmoothingMiddle := StrToProportionFloat(RValue, 1/100, 99/100);
							if bLogging and bLogConfigFile then LogCall('[ADC] SmoothingMiddle = ' + FloatToStr(100 * Dev.SmoothingMiddle) + '%.');
						except
						end
						else if LValue = 'SMOOTHINGREPS' then try
							Dev.SmoothingASDReps := Max(1, StrToInt(RValue));
							if bLogging and bLogConfigFile then LogCall('[ADC] SmoothingReps = ' + IntToStr(Dev.SmoothingASDReps) + '.');
						except
						end
						else if LValue = 'SMOOTHINGINERTIA' then try
							Dev.SmoothingInertia := Min(192, StrToInt(RValue));
							if bLogging and bLogConfigFile then LogCall('[ADC] SmoothingInertia = ' + IntToStr(Dev.SmoothingInertia) + '.');
						except
						end
						else begin
							if bLogging and bLogConfigFile then LogCall('Unrecognized [ADC] L-value "' + LValue + '".');
						end;
					end
					else if ThisSection = '[DAC]' then begin
						if LValue = 'CAL0M' then try
							ConfigDACCal.M := StrToFloat(RValue);
							if bLogging and bLogConfigFile then LogCall('[DAC] Cal0M = ' + FloatToStr(ConfigDACCal.M) + '.');
						except
						end
						else if LValue = 'CAL0B' then try
							ParseVStrAsB(ConfigDACCal.B, DACCalScale, RValue);
							if bLogging and bLogConfigFile then LogCall('[DAC] Cal0B = ' + FloatToStr(ConfigDACCal.B) + '.');
						except
						end
						else begin
							if bLogging and bLogConfigFile then LogCall('Unrecognized [DAC] L-value "' + LValue + '".');
						end;
					end
					else begin
						if bLogging and bLogConfigFile then LogCall('Command "' + Content + '" in unrecognized section.');
					end;
				end
				else begin
					if bLogging and bLogConfigFile then LogCall('Intransitive command "' + Content + '".');
				end;
			end;
		except
			on E: Exception do if bLogging then LogCall(E.Message + ': ' + E.ClassName);
		end;
		Contents.Free;
	end
	else begin
		if bLogging and bLogConfigFile then LogCall('No config file found.');
	end;
	if bLogging and bLogConfigFile then begin
		LogCall('Done with config file loading.');

		LogCall('A/D Cal: VM = ' + FloatToStr(ConfigADCal[False].M) + ', VB = ' + FloatToStr(ConfigADCal[False].B) + ', IM = ' + FloatToStr(ConfigADCal[True].M) + ', IB = ' + FloatToStr(ConfigADCal[True].B) + '.');
	end;
end;



function DoBCControl(BCControlData: LongWord; TCSamples: LongWord = 0): LongWord;
begin
	Result := GenericVendorWrite(DeviceIndex, $BC, 0, TCSamples, SizeOf(BCControlData), @BCControlData);
end;

function AWU_GenericBulkIn(DeviceIndex: LongWord; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesRead: LongWord): LongWord; cdecl; external 'AIOUSB.dll';

function ADC_BulkAcquireBlocking(DeviceIndex: LongWord; BufSize: LongWord; pBuf: Pointer): LongWord; cdecl;
var
	BytesLeft: Integer;
	BytesRead: LongWord;
	pDataBuf: PByte;

begin
	BytesLeft := BufSize;
	pDataBuf := pBuf;

	DoBCControl($00000005, BufSize div 2); //We actuate the read and do the read from the same thread, so this can't be generic.
	ADC_GetImmediate(DeviceIndex, 0, pBuf);
	repeat
		BytesRead := 0;
		Result := AWU_GenericBulkIn(DeviceIndex, $86, pDataBuf, BytesLeft, BytesRead);
		if Result <> ERROR_SUCCESS then Break;
		Dec(BytesLeft, BytesRead);
		Inc(pDataBuf, BytesRead);
	until BytesLeft <= 0;
end;





	//Opens handles, ports, whatever, for use.

function di_open(USBID: LongWord): Word; stdcall;
var
	I: Integer;
	ConfigSize: LongWord;
	OutMask: Byte;
	DIOData: Word;

begin
	if DeviceIndex <> diNone then begin
		Result := DI_ALREADY_OPEN_ERR;
		Exit;
	end;

	hCardMutex := CreateMutex(nil, False, CardMutexNom);
	WaitForSingleObject(hCardMutex, INFINITE);

	//Reload device links. (This will interrupt everything in progress, so it'd have to change for multiple simultaneous devices.)
	for I := 0 to 31 do
		AIOUSB_CloseDevice(I)
	;
	GetDevices;

	DACDeviceIndex := diNone;
	DeviceIndex := GetDeviceByEEPROMByte(USBID);
	if DeviceIndex = diNone then begin
		Result := DI_DRIVER_ERR;
		ReleaseMutex(hCardMutex);
		CloseHandle(hCardMutex);
		Exit;
	end;

	DACDeviceIndex := GetDeviceByEEPROMByte(USBID or $80);
	if DACDeviceIndex = diNone then begin
		DACDeviceIndex := DeviceIndex;
		DACBase := 0;
	end
	else begin
		DACBase := 3;
	end;
	{
	if DACDeviceIndex = diNone then begin
		DeviceIndex := diNone; 
		Result := DI_DRIVER_ERR;
		ReleaseMutex(hCardMutex);
		CloseHandle(hCardMutex);
		Exit;
	end;
	}

	QueryPerformanceFrequency(TicksPerSec);
	if bLogging then begin
		MicrosPerTick := 1000000 / TicksPerSec;
		InitializeLogging(bDateTimeLog);
		
		if bLogCalls then LogCall('di_open');
	end;

	SetString(ConfigFileName, nil, MAX_PATH);
	I := GetModuleFileName(0, @ConfigFileName[1], Length(ConfigFileName));
	SetLength(ConfigFileName, I);
	ConfigFileName := IncludeTrailingBackslash(ExtractFilePath(ConfigFileName)) + 'USBAIODICfg.ini';
	DoLoadConfigFile;

	with Dev do begin
		//Low byte is output, high byte is input.
		OutMask := $1;
		DIOData := $FFFF;
		DIO_Configure(DeviceIndex, False, @OutMask, @DIOData);

		if DACBase = 0 then
			for I := 0 to 1 do
				DACDirect(DACDeviceIndex, I, $8000)
		else
			for I := 0 to 7 do
				DACDirect(DACDeviceIndex, I, $8000)
		;
		DACSetBoardRange(DACDeviceIndex, 3);
		DACLastOut[0] := -1;
		DACLastOut[1] := -1;
		DACOut(0, 0);
		DACOut(1, 0);

		ConfigSize := Length(ShadowConfig);
		ADC_GetConfig(DeviceIndex, @ShadowConfig[0], ConfigSize);
		for I := $0 to $F do ShadowConfig[I] := $1;
		ShadowConfig[$11] := $15;
		SetConfig;

		ADC_SetCal(DeviceIndex, ':1TO1:');

		ElectSerialNumber(DeviceIndex, SN);
		NumChannels := 16; //Note, ChannelShift is not implemented.
		SamplesPerChannel := 3;
		ScanHz := 0;
		BurstHz := 40000;
		InListLen := 0;
		OutListLen := 0;
		with AIBuf do if hIndexMutex = 0 then hIndexMutex := CreateMutex(nil, False, nil);
		with AOBuf do if hIndexMutex = 0 then hIndexMutex := CreateMutex(nil, False, nil);
		if hDACLastMutex = 0 then hDACLastMutex := CreateMutex(nil, False, nil);
		bScanning := False;
		bSuspendedByMisuse := False;
		bRejectingData := False;
		bAOStopped := False;
		bWriteHead := True;
		DACBufferStuffed := 0;
		LastDACReport := 0;
		SetLength(Hist, $10000);
		EID := USBID;
	end;

	for I := 0 to High(RangeCal) do with RangeCal[I] do begin
		M := 1;
		B := 0;
	end;

	Result := DI_NO_ERR;
end;

	//Ends all running sequences and closes communication ports/handles/etc.

function di_close(): Word; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_close');

	//Check if streams are running, if so end them.
	di_stop_scan;

	AIOUSB_CloseDevice(DeviceIndex);
	DeviceIndex := diNone;

	ReleaseMutex(hCardMutex);
	CloseHandle(hCardMutex);

	if bLogging then FinalizeLogging;

	Result := DI_NO_ERR;
end;

function AIO_AddCal(CalVB, CalIB: Double): LongWord; stdcall;
var
	Content, LValue, RValue, ThisSection: AnsiString;
	Contents: TStringList;
	iCalB, iCalVB, iCalIB, iLastADC, iLine, I: Integer;
	CalStrV, CalStrI: AnsiString;

begin
	if bLogging and bLogCalls then LogCall('AIO_AddCal(' + Modif(CalVB) + ', ' + Modif(CalIB) + ')');

	Contents := TStringList.Create;
	try
		if FileExists(ConfigFileName) then try
			Contents.LoadFromFile(ConfigFileName);
		except
			Contents.Clear;
		end;

		iLastADC := -1;
		iCalB := -1;
		iCalVB := -1;
		iCalIB := -1;
		CalStrV := 'CalVB = ' + FloatToStr(1000/CountsPerVolt/ADCCalScale * ConfigADCal[False].B + CalVB);
		CalStrI := 'CalIB = ' + FloatToStr(1000/CountsPerVolt/ADCCalScale * ConfigADCal[True ].B + CalIB);

		if bLogging then begin
			LogCall('CalVB = ' + FloatToStr(1000/CountsPerVolt/ADCCalScale * ConfigADCal[False].B) + Modif(CalVB));
			LogCall('CalIB = ' + FloatToStr(1000/CountsPerVolt/ADCCalScale * ConfigADCal[True ].B) + Modif(CalIB));
		end;

		ThisSection := '';
		for iLine := 0 to Contents.Count - 1 do begin
			Content := Trim(Contents[iLine]);
			if Content = '' then Continue;
			if Content[1] = ';' then begin
				if ThisSection = '[ADC]' then iLastADC := iLine;
				Continue;
			end;

			if (Content[1] = '[') and (Content[Length(Content)] = ']') then begin
				//Sections.
				ThisSection := Content;
				Continue;
			end;

			I := Pos('=', Content);
			if I <> 0 then begin
				//Assign commands.
				LValue := UpperCase(TrimRight(Copy(Content, 1, I-1)));
				RValue := TrimLeft(Copy(Content, I+1, MAXINT));

				if ThisSection = '[ADC]' then begin
					iLastADC := iLine;
					if LValue = 'CALB' then
						iCalB := iLine
					else if LValue = 'CALVB' then
						iCalVB := iLine
					else if LValue = 'CALIB' then
						iCalIB := iLine
					;
				end;
			end;
		end;

		if iLastADC < 0 then begin
			Contents.Add('');
			Contents.Add('[ADC]');
			Contents.Add(CalStrV);
			Contents.Add(CalStrI);
		end
		else begin
			if (iCalVB >= 0) and (iCalVB > iCalB) then begin
				Contents[iCalVB] := CalStrV;
			end
			else begin
				Inc(iLastADC);
				Contents.Insert(iLastADC, CalStrV);
			end;

			if (iCalIB >= 0) and (iCalIB > iCalB) then begin
				Contents[iCalIB] := CalStrI;
			end
			else begin
				Inc(iLastADC);
				Contents.Insert(iLastADC, CalStrI);
			end;
		end;

		DeleteFile(ConfigFileName);
		Sleep(1); //NTFS has problems if we try to touch a file right after deleting it.
		Contents.SaveToFile(ConfigFileName);
		Result := ERROR_SUCCESS;
	except
		Result := ERROR_INTERNAL_ERROR;
	end;
	Contents.Free;

	if Result = ERROR_SUCCESS then DoLoadConfigFile;
end;





//Note that this gets dumped to the screen, and may therefore need more.
function di_info(var A1: TDIInfoData16): Word; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_info');

	ZeroMemory(@A1, SizeOf(A1));
	A1.serial_no := Dev.SN and $7FFFFFFF;
	A1.board_id := '720COMPAT'#0;

	Result := DI_NO_ERR;
end;

	//Same as di_info, but 32-bit aligned: "di_info32 has padding to align 32-bit structure members declared in HDRDEF32.H on 32-bit boundaries".

function di_info32(var A1: TDIInfoData32): Word; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_info32');

	ZeroMemory(@A1, SizeOf(A1));
	A1.serial_no := Dev.SN and $7FFFFFFF;
	A1.board_id := '720COMPAT'#0;

	Result := DI_NO_ERR;
end;





	//Reads counts for one input channel(at a selectable gain/range/polarity).

function di_anin(const AninData: TDIAnInData): SmallInt; stdcall;
var
	RangeCode: Byte;
	Scan: array of Word;

begin
	if DeviceIndex = diNone then begin
		Result := 0;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_anin');

	if Dev.bScanning then begin
		if bLogging then LogCall('While scanning!');
		Result := 0;
		Exit;
	end;

	with Dev do begin
		Result := 0;
		if AninData.chan >= NumChannels then Exit;

		RangeCode := DIGainCodeToRangeCode(AninData.Gain);
		if AninData.unipolar <> 0 then RangeCode := RangeCode and not 1;
		if (AninData.diff <> 0) and (AninData.chan < 8) then RangeCode := RangeCode or 8;

		ShadowConfig[$12] := $11 * AninData.chan;
		ShadowConfig[AninData.chan] := RangeCode;
		ShadowConfig[$13] := $F;
		SetConfig;

		SetLength(Scan, NumChannels);
		ADC_GetScan(DeviceIndex, @Scan[0]);
		Result := Scan[AninData.chan] xor $8000;
	end;
end;

	//Writes one analog output value.

function di_anout(DACData: SmallInt; Channel: Word): Word; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_anout(' + IntToHex(DACData, 4) + ', ' + IntToHex(Channel, 1) + ')');

	DACOut(Channel, DACData);

	if Dev.bSuspendedByMisuse then begin
		//This is the last distinctive call they make when doing a bunch of config in the middle of a scan.
		//Since we suspend for doing things in the middle of a scan, resume at this point.
		Dev.bSuspendedByMisuse := False;
		di_start_scan;
	end;

	Result := DI_NO_ERR;
end;





	//Reads and returns two bytes of digital data.

function di_digin(): Word; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := 0;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_digin');

	DIO_ReadAll(DeviceIndex, @Result);
end;

	//Writes two bytes of digital data. Bits default to "high".

function di_digout(DIOData: Word): Word; stdcall;
var
	I: Integer;


begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_digout(' + IntToHex(DIOData, 2) + ')');

	if DebugDelay >= 0 then Sleep(DebugDelay);
	for I := 0 to HammerExtra do DIO_WriteAll(DeviceIndex, @DIOData);

	Result := DI_NO_ERR;
end;





	//Allocates the driver's internal ring buffers.
	//Could take this time to allocate a buffer above the cont callback, if that's easier.
	//Returns pointer to the buffer allocated.

function di_buffer_alloc(Dir, BufferWords: Word): PWord; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := nil;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_buffer_alloc(' + DirToStr[Dir] + ', ' + IntToHex(BufferWords, 1) + ')');

	if Dev.bScanning then begin
		if bLogging then LogCall('While scanning!');

		case Dir of
			0: with Dev.AIBuf do begin
				WaitForSingleObject(hIndexMutex, INFINITE);
					NextIndex := 0;
					Dev.bWriteHead := True; 
				ReleaseMutex(hIndexMutex);
			end;
			1: with Dev.AOBuf do begin
				WaitForSingleObject(hIndexMutex, INFINITE);
					NextIndex := 0;
					Dev.DACBufferStuffed := BufferWords;
				ReleaseMutex(hIndexMutex);
			end;
		end;

		Result := nil;
		Exit;
	end;

	case Dir of
		0: with Dev.AIBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				SetLength(Buf, 0);
				SetLength(Buf, 2*BufferWords); //Allocate double the buffer - it's small, and we do pass back the pointer, after all.
				BufLen := BufferWords;
				Result := @Buf[0];
				ZeroMemory(Result, 2*BufferWords);
				NextIndex := 0;
				Dev.bWriteHead := True;
			ReleaseMutex(hIndexMutex);
		end;
		1: with Dev.AOBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				SetLength(Buf, 0);
				SetLength(Buf, 2*BufferWords); //Allocate double the buffer - it's small, and we do pass back the pointer, after all.
				BufLen := BufferWords;
				Result := @Buf[0];
				ZeroMemory(Result, 2*BufferWords);
				NextIndex := 0;
				Dev.DACBufferStuffed := BufferWords;
			ReleaseMutex(hIndexMutex);
		end;
		else begin
			Result := nil;
		end;
	end;
end;

	//Deallocates the driver's internal ring buffers.
	//Could take this time to deallocate the buffer, as above.

function di_buffer_free(Dir: Word): Word; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging then begin
		if bLogCalls then LogCall('di_buffer_free(' + DirToStr[Dir] + ')');

		if FinalDataLog <> fdlNone then
			if Dir = 1 then
				FinalizeFDLDACCat
		;
	end;

	if Dev.bScanning then begin
		if bLogging then LogCall('While scanning!');

		//This is the first distinctive call they make when doing a bunch of config in the middle of a scan.
		//Suspend at this point.
		Dev.bSuspendedByMisuse := True;
		di_stop_scan;

		Result := DI_COMM_ERR;
		Exit;
	end;

	case Dir of
		0: with Dev.AIBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				SetLength(Buf, 0);
				NextIndex := 0;
				BufLen := 0;
			ReleaseMutex(hIndexMutex);
		end;
		1: with Dev.AOBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				SetLength(Buf, 0);
				NextIndex := 0;
				BufLen := 0;
				Dev.DACBufferStuffed := 0;
			ReleaseMutex(hIndexMutex);
		end;
	end;

	Result := DI_NO_ERR;
end;





	//This function allocates driver memory internal to hold the point lists initialized by di_inlist and di_outlist.

function di_list_length(InLength, OutLength: Word): Word; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_list_length(' + IntToHex(InLength, 1) + ', ' + IntToHex(OutLength, 1) + ')');

	if Dev.bScanning then begin
		if bLogging then LogCall('While scanning!');

		with Dev.AIBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				NextIndex := 0;
				Dev.bWriteHead := True;
			ReleaseMutex(hIndexMutex);
		end;
		with Dev.AOBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				NextIndex := 0;
			ReleaseMutex(hIndexMutex);
		end;

		Result := DI_COMM_ERR;
		Exit;
	end;

	Dev.InListLen := InLength;
	Dev.OutListLen := OutLength;

	if bLoadConfigOnListLen then DoLoadConfigFile;

	Result := DI_NO_ERR;
end;

	//Configure ADC channels to scan, with optional oversampling, and per-channel gain etc. Doesn't start until di_start_scan. Won't be used for digital output.
	//A bit hacky, we know they pass in all the channels, albeit in a strange order.

function di_inlist(pEntry: TPDIInlistData): Word; stdcall;
const
	DIGainCodeToGainFactor: array[0..3] of Integer = (1, 2, 5, 10);

var
	AICh, iTissue, I: Integer;
	RangeCode: Byte;
	bCurrent: Boolean;

begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging then begin
		if bLogCalls then LogCall('di_inlist');

		//GainBuf := 'Gains:';
		if FinalDataLog <> fdlNone then SetLength(GainFactor, Dev.NumChannels);
	end;

	if Dev.bScanning then begin
		if bLogging then LogCall('While scanning!');

		with Dev.AIBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				NextIndex := 0;
				Dev.bWriteHead := True;
			ReleaseMutex(hIndexMutex);
		end;

		Result := DI_COMM_ERR;
		Exit;
	end;

	Sleep(1);

	Dev.TheirSamplesPerChannel := 1 + pEntry.counter; //They happen to set them all equal, so use the first entry in the pointlist.
	if bLogging and bLogStats then LogCall('Dev.TheirSamplesPerChannel: ' + IntToHex(Dev.TheirSamplesPerChannel, 1));
	for I := 0 to Dev.InListLen - 1 do begin
		if pEntry.Chan >= Dev.NumChannels then Continue;

		if bLogging then if FinalDataLog <> fdlNone then if I < Dev.NumChannels then GainFactor[I] := DIGainCodeToGainFactor[pEntry.Gain];
		RangeCode := DIGainCodeToRangeCode(pEntry.Gain);
		bCurrent := pEntry.Chan >= $8;
		iTissue := pEntry.Chan and $7;
		AICh := (iTissue shl 1) or Ord(bCurrent);
		if (pEntry.Diff <> 0) and (AICh < 8) then RangeCode := RangeCode or 8;
		Dev.ShadowConfig[AICh] := RangeCode; //Note, ChannelShift is not implemented.
		//pEntry.dig_out_enable is ignored, DIO is not done in a stream.
		//pEntry.Ave is ignored, our idiom is always some last-point and some averaging.
		//pEntry.counter is ignored, except for the first entry's affecting TheirSamplesPerChannel.

		if bLogging then if 1 + pEntry.counter <> Dev.TheirSamplesPerChannel then LogCall(' <> ' + IntToHex(1 + pEntry.counter, 1) + ' @ ' + IntToHex(I, 1));

		Inc(pEntry);
	end;

	with Dev do begin
		ShadowConfig[$12] := $F0; //Note: This affects the simplified channel count calculation near the end of di_start_scan.
		SetSpeeds(True); //Calls SetConfig.

		if bLogging then begin
			if FinalDataLog <> fdlNone then DoLogGains;
		end;
	end;

	Result := DI_NO_ERR;
end;

	//Similar to DACOutputProcess, but in the .DLL. Doesn't start until di_start_scan. Won't be used for digital output.
	//A bit hacky, we know they only pass in one entry.

function di_outlist(pEntry: TPDIOutlistData): Word; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_outlist');

	if Dev.bScanning then begin
		if bLogging then LogCall('While scanning!');

		with Dev.AOBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				NextIndex := 0;
			ReleaseMutex(hIndexMutex);
		end;

		Result := DI_COMM_ERR;
		Exit;
	end;

	if bLogging and bLogStats then LogCall('DAC counter: ' + IntToHex(pEntry.counter));
	with Dev do begin
		DACChannel := pEntry.Unipolar;
		//pEntry.digital is ignored, they don't use it.
	end;

	Result := DI_NO_ERR;
end;

function di_burst_rate(Divisor: Word): Word; stdcall;
begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_burst_rate(' + IntToHex(Divisor, 1) + ')');

	if Dev.bScanning then begin
		if bLogging then LogCall('While scanning!');
		Result := DI_COMM_ERR;
		Exit;
	end;

	with Dev do begin
		BurstHz := 16000000/Divisor;

		SetSpeeds;
	end;

	Result := DI_NO_ERR;
end;





	//Run the configured process(di_inlist and/or di_outlist).

function di_start_scan(): Word; stdcall;
var
	ScanSize, BufSize: LongWord;
	OutCh, I: Integer;
	RangeCode: Byte;
	bCurrent: Boolean;
	BufCount: Integer;

begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if Dev.bScanning then begin
		Result := DI_START_SCN_ERR;
		Exit;
	end;

	if Dev.ScanHz = 0 then begin
		Result := DI_MODE_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_start_scan');

	with Dev do begin
		FreqOut(0);
		SetConfig;
		ThisAICh := 0;
		ThisAIOversample := 0;

		IncompleteChannel := 0;
		for I := $0000 to $FFFF do Hist[I] := 0;
		I := SamplesPerChannel - DiscardOversamples;
		iSmoothingCull := Min((I-1) div 2, Max(1, Round(I * (1 - SmoothingMiddle)/2))); //Instead of applying a max and min, we could fall back to smAverage or smMedian, but that could destroy information.

		SetLength(IncompleteScan, NumChannels);
		for I := 0 to $F do IncompleteScan[I] := 0;
		SetLength(ChSaturation, NumChannels);
		//Ignoring the mutexes because the other threads aren't launched yet.
		DACLastOut[0] := -1;
		DACLastOut[1] := -1;
		AIBuf.NextIndex := 0;
		AOBuf.NextIndex := 0;
		LogCall('AIOUSB_ClearFIFO(5)');
		AIOUSB_ClearFIFO(DeviceIndex, 5);
		LastBufferStatus := -1;
		bAOStopped := False;
		bWriteHead := True;
		nData := $FFFF;
		xData := $0000;
		for I := 0 to $F do OuterDataOriginal[I] := 0;
		bSaturatedPulse := False;
		AccScanStr := '';
		RawScanStr := '';
		for I := $0 to $F do begin
			RangeCode := ShadowConfig[I] and $7;
			OutCh := I;
			bCurrent := (OutCh and 1) <> 0;
			OutChCal[OutCh].M := ConfigADCal[bCurrent].M * RangeCal[RangeCode].M;
			OutChCal[OutCh].B := ConfigADCal[bCurrent].M * RangeCal[RangeCode].B + ConfigADCal[bCurrent].B * ChGainFactor[I];
		end;

		if bLogging then begin
			//LogCall(IntToHex(Round(ScanHz), 1) + ' = ' + IntToHex(Round(BurstHz), 1) + ' / (' + IntToHex(TheirSamplesPerChannel, 1) + ' * ' + IntToHex(Max(InListLen, OutListLen), 1) + ')');
			LogCall('Scanning at ' + IntToHex(Round(ScanHz), 1) + '.');
		end;

		//`@ Use many single-scan buffers for responsiveness.
		BufSize := 8*512;
		BufCount := 512;

		ScanSize := 2 * (1 + ShadowConfig[$13]) * (1 + (ShadowConfig[$12] shr 4)); //Simplified channel count calculation.
		TicksPer3Callbacks := Round(3 * ScanSize * TicksPerSec
		                              / (BufSize * ScanHz)
		);
		QueryPerformanceCounter(LastCallbackTick);
		bGotCallback := False;

		ADC_BulkContinuousCallbackStart(DeviceIndex, BufSize, BufCount, 0, @ADContCallback);

		if DebugDelay >= 0 then Sleep(DebugDelay);
		FreqOut(ScanHz);

		bScanning := True;
	end;

	Result := DI_NO_ERR;
end;

	//Stop any scans in process.
	//Note: Called internally by di_close.

function di_stop_scan(): Word; stdcall;
var
	IOStatus: LongWord;
	MyAbortThread: TAbortThread;

begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if not Dev.bScanning then begin
		Result := DI_STOP_SCN_ERR;
		Exit;
	end;

	if bLogging and bLogCalls then LogCall('di_stop_scan');

	with Dev do begin
		bRejectingData := True;

		DoBCControl($01000007);

		if bLogging then LogCall('ADC_BulkContinuousEnd');
		MyAbortThread := TAbortThread.BeginAbortWait(EID);
		ADC_BulkContinuousEnd(DeviceIndex, @IOStatus);
		MyAbortThread.EndAbortWaitAndFree;
		FreqOut(0);
		SetConfig;

		bRejectingData := False;
		bScanning := False;
	end;

	Result := DI_NO_ERR;
end;

function RestartPulse: Boolean;
begin
	Result := WaitForSingleObject(Dev.AIBuf.hIndexMutex, 20) = WAIT_OBJECT_0;
	if not Result then Exit;

		Dev.AIBuf.NextIndex := 0;
		Dev.bWriteHead := True;
		Dev.bAOStopped := False;
	ReleaseMutex(Dev.AIBuf.hIndexMutex);
end;

	//Returns how much data is in the buffer.

function di_buffer_status(Dir: Word): LongWord; stdcall;
var
	Naui: Int64;

begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging then LogThreadLogCalls;

	if Dev.bScanning and Dev.bGotCallback then begin
		QueryPerformanceCounter(Naui);
		Naui := Naui - Dev.LastCallbackTick;
		if Naui > Dev.TicksPer3Callbacks then begin
			Dev.bGotCallback := False;
			if bLogging then LogCall('Watchdog Timeout  It''s been ' + FloatToStr(1000 * Naui/TicksPerSec) + 'ms since we got a callback, but we should be getting a callback every ' + FloatToStr(1000/3 * Dev.TicksPer3Callbacks/TicksPerSec) + 'ms.');
			di_stop_scan;
			di_start_scan;
		end;
	end;

	case Dir of
		0: with Dev.AIBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				Result := NextIndex;
			ReleaseMutex(hIndexMutex);
		end;
		1: begin
			with Dev.AOBuf do begin
				WaitForSingleObject(hIndexMutex, INFINITE);
					if (NextIndex >= Dev.DACBufferStuffed) and (NextIndex = Dev.LastDACReport) then begin
						Inc(NextIndex);
						if NextIndex >= BufLen then begin
							NextIndex := 0;
							RestartPulse;
						end;
					end;
					Result := NextIndex;
				ReleaseMutex(hIndexMutex);
			end;

			Dev.LastDACReport := Result;
		end;
		else begin
			Result := 0;
		end;
	end;

	if bLogging then begin
		//QueryPerformanceCounter(Naui);
		//if BufferStatusLog = bslAlways then LogCall('di_buffer_status(' + DirToStr[Dir] + ') = ' + IntToHex(Result, 1) + ' / ' + IntToHex(L, 1) + ' (' + IntToStr(Round(MicrosPerTick * (Naui - Den))) + 's)');
		if BufferStatusLog = bslAlways then LogCall('di_buffer_status(' + DirToStr[Dir] + ') = ' + IntToHex(Result, 1));
		if (Dir = 1) and (Integer(Result) <> Dev.LastBufferStatus) then begin
			Dev.LastBufferStatus := Result;
			if BufferStatusLog = bslOnChange then LogCall('di_buffer_status(' + DirToStr[Dir] + ') = ' + IntToHex(Result, 1));
		end;
	end;
end;

procedure PushForAutorange(pDest: Pointer; Count: Word);
var
	OldSample, OuterDataFinal, iOuterDataFinal: Integer;
	pThisDest: PWord;
	iCh, iMappedCh, hScan, iScan: Integer;
	D, M: Integer;

begin
	if not Dev.bSaturatedPulse then begin
		hScan := Min(Dev.LastBufferStatus, Count div Dev.NumChannels)-1;
		for iCh := 0 to Dev.NumChannels-1 do begin
			iMappedCh := iCh;

			//Find the most extreme sample in the block.
			OuterDataFinal := 0;
			iOuterDataFinal := -1;
			pThisDest := pDest;
			Inc(pThisDest, iMappedCh + Dev.NumChannels);
			for iScan := 1 to hScan do begin
				if (LastDACCat[iScan-1] <> 0) and (LastDACSection[iScan] = LastDACSection[iScan-1]) then begin
					M := Abs(PSmallInt(pThisDest)^);
					if OuterDataFinal < M then begin
						OuterDataFinal := M;
						iOuterDataFinal := iScan;
					end;
				end;
				Inc(pThisDest, Dev.NumChannels);
			end;

			//If the most extreme raw value is more extreme, then bump it out.
			if (iOuterDataFinal >= 0) and (Dev.OuterDataOriginal[iCh] > OuterDataFinal) then begin //Only accept samples that're not in the center DAC value, and that aren't first in their section.
				D := Dev.OuterDataOriginal[iCh] - OuterDataFinal;
				pThisDest := pDest;
				Inc(pThisDest, iMappedCh + Dev.NumChannels * iOuterDataFinal);
				OldSample := PSmallInt(pThisDest)^;
				if OldSample < 0 then D := -D;
				//Make sure we won't try to compensate for a value that's past saturation.
				if (OldSample + D) > $7FFF then
					D := $7FFF - OldSample
				else if (OldSample + D) < -$8000 then
					D := -$8000 - OldSample
				;
				pThisDest^ := OldSample + D; //Store a value with magnitude at least that extreme.

				//Find another value of the same section on the same channel, preferrably after(because before might be the reject).
				if LastDACSection[iOuterDataFinal+1] = LastDACSection[iOuterDataFinal] then
					Inc(pThisDest, Dev.NumChannels)
				else
					Dec(pThisDest, Dev.NumChannels)
				;
				//Push it the opposite way to compensate. The average should stay the same.
				pThisDest^ := Min($7FFF, Max(-$8000, Integer(PSmallInt(pThisDest)^) - D));
			end;
		end;
	end;
end;

	//Pull data from driver's input buffer to user array.

function di_copy_buffer(BufIndex: Word; pDest: Pointer; Count: Word): Word; stdcall;
begin
	try
		if DeviceIndex = diNone then begin
			Result := DI_OPENED_ERR;
			Exit;
		end;

		if bLogging and bLogCalls then LogCall('di_copy_buffer(' + IntToHex(BufIndex, 1) + ', ' + IntToHex(Count, 1) + ')');

		with Dev.AIBuf do begin
			WaitForSingleObject(hIndexMutex, INFINITE);
				Move(Buf[BufIndex], pDest^, 2*Count);
			ReleaseMutex(hIndexMutex);
		end;

		PushForAutorange(pDest, Count);

		if bLogging then case FinalDataLog of
			fdlRaw       : LogRawData       (Min(Dev.NumChannels * Dev.LastBufferStatus, Count), pDest);
			fdlRawDecimal: LogDecimalRawData(Min(Dev.NumChannels * Dev.LastBufferStatus, Count), pDest);
			fdlParsed    : LogParsedData    (Min(Dev.NumChannels * Dev.LastBufferStatus, Count), pDest, Dev.NumChannels);
		end;

		Result := DI_NO_ERR;
	except
		on E: Exception do begin
			Result := DI_DRIVER_ERR;
		end;
	end;
end;

	//Push data from user array to driver's output buffer, by address.

function di_copy_array(BufIndex: Word; pSource: PWord; Count: Word): Word; stdcall;
var
	BaseData: Word;
	I: Integer;

begin
	if DeviceIndex = diNone then begin
		Result := DI_OPENED_ERR;
		Exit;
	end;

	if bLogging then begin
		if bLogCalls then LogCall('di_copy_array(' + IntToHex(BufIndex, 1) + ', ' + IntToHex(Count, 1) + ')');
	end;

	SetupFromDACData(pSource, Count, bLogStats);

	with Dev.AOBuf do begin
		WaitForSingleObject(hIndexMutex, INFINITE);
			Move(pSource^, Buf[BufIndex], 2*Count);
			if BufIndex = 0 then NextIndex := 0;
			Dev.bAOStopped := False;
			if Dev.DACBufferStuffed = BufLen then
				Dev.DACBufferStuffed := BufIndex + Count
			else
				Dev.DACBufferStuffed := Max(Dev.DACBufferStuffed, BufIndex + Count)
			;

			BaseData := Buf[BufIndex];
			Dev.BeginningOfTheEnd := BufIndex;
			for I := Integer(BufIndex) + Integer(Count) - 1 downto BufIndex do
				if Buf[I] <> BaseData then begin
					Dev.BeginningOfTheEnd := I+1;
					Break;
				end
			;
		ReleaseMutex(hIndexMutex);
	end;

	Result := DI_NO_ERR;
end;





procedure TickFastDAC;
var
	ThisSample: SmallInt;
	FromIndex: Integer;

begin
	if Dev.bAOStopped then Exit;

	with Dev.AOBuf do begin
		WaitForSingleObject(hIndexMutex, INFINITE);
			FromIndex := NextIndex;
			ThisSample := Buf[NextIndex];
			Inc(NextIndex);
			if NextIndex >= BufLen then begin
				NextIndex := 0;
				if not RestartPulse then Dev.bAOStopped := True;
			end;
		ReleaseMutex(hIndexMutex);
	end;

	if FromIndex >= Dev.DACBufferStuffed then Exit;

	DACOut(Dev.DACChannel, ThisSample, True);
end;

procedure PostAIScan(pDev: TPDeviceData);
var
	L, BufCt, NextI, nNextI: Integer;

begin
	with pDev^ do begin
		WaitForSingleObject(AIBuf.hIndexMutex, INFINITE);
			L := AIBuf.BufLen;
			NextI := AIBuf.NextIndex;
			nNextI := NextI + NumChannels;
			if nNextI >= L then bWriteHead := False;
			if nNextI > L then begin
				//Wraps, so split into two buffers. First buffer goes at the end, second buffer goes at the beginning.
				if bLogging then ThreadLogCall('That''s a wrap!');
				BufCt := L - NextI;
				if bWriteHead then Move(IncompleteScan[0], AIBuf.Buf[NextI], 2*BufCt);
				nNextI := nNextI - L;
				if bWriteHead then Move(IncompleteScan[BufCt], AIBuf.Buf[0], 2*nNextI);
			end
			else begin
				//Fits as one buffer.
				if bWriteHead then Move(IncompleteScan[0], AIBuf.Buf[NextI], 2*NumChannels);
				if nNextI = L then nNextI := 0;
			end;
			AIBuf.NextIndex := nNextI;
		ReleaseMutex(AIBuf.hIndexMutex);
	end;
end;

function BalanceAsymmetricStandardDeviations(PrevApprox: Double; MethodReps: Integer = 1): Double;
var
	VarianceL, VarianceH, NextApprox: Double;
	WL, WH, I, J: Integer;

begin
	Result := PrevApprox;
	with Dev do
		for J := 1 to MethodReps do begin
			//Split the data points into "below average" and "above average", calculating a variance for each.
			VarianceL := 0; WL := 0;
			for I := nData to Floor(Result) do begin
				VarianceL := VarianceL + Hist[I] * Sqr(I - Result);
				Inc(WL, Hist[I]);
			end;
			VarianceH := 0; WH := 0;
			for I := Ceil(Result) to xData do begin
				VarianceH := VarianceH + Hist[I] * Sqr(I - Result);
				Inc(WH, Hist[I]);
			end;
			//Turn the variances into standard deviations, and bias the average toward the larger.
			if (WL <> 0) and (WH <> 0) then begin
				NextApprox := Result + 1/2 * (SqRt(VarianceH/WH) - SqRt(VarianceL/WL));
				if (NextApprox <= nData) or (NextApprox >= xData) then begin
					if bLogging then ThreadLogCall('Warning: Bailed out on rep ' + IntToHex(J) + '/' + IntToHex(MethodReps) + ' from BalanceAsymmetricStandardDeviations. ' + IntToHex(nData) + '..' + IntToHex(Round(NextApprox)) + '..' + IntToHex(xData) + '.');
					Exit;
				end;
				Result := NextApprox;
			end
			else begin
				if bLogging then ThreadLogCall('Warning: Bailed out on rep ' + IntToHex(J) + '/' + IntToHex(MethodReps) + ' from BalanceAsymmetricStandardDeviations. ' + IntToHex(WH) + ' above, ' + IntToHex(WL) + ' below.');
				Exit; //Something very strange has happened to the standard deviations, so stick with the previous approximation.
			end;
		end
	;
end;

procedure ADContCallback(pBuf: PWord; BufSize, Flags, Context: LongWord);
var
	BufCt: Integer;
	n2, x2, iRawSample, iScanSample, I, W: Integer;
	ThisNXScanStr: AnsiString;
	ThisRawScanStr: AnsiString;
	ThisSample: Double;
	ThisSample16: SmallInt;
	Saturate: (stNone, stLower, stUpper, stNearest);

begin
	BufCt := BufSize div 2;
	if (BufCt = 0) or (pBuf = nil) then Exit; //We don't care about Flags.

	//if bLogging then ThreadLogCall('ADContCallback(' + IntToHex(BufSize, 1) + ', ' + IntToHex(Flags, 1) + ')');

	QueryPerformanceCounter(Dev.LastCallbackTick);
	Dev.bGotCallback := True;

	with Dev do begin
		if ThisAICh + ((ThisAIOversample + BufCt) div SamplesPerChannel) >= NumChannels then
			TickFastDAC //This includes the end of a scan, so tick the DAC.
		;

		for iRawSample := 1 to BufCt do begin
			if bLogging and bLogCallbackDecimalRaw then begin
					if ThisAIOversample = 0 then
						RawScanStr := RawScanStr + (#13#10 + LastDACSection[AIBuf.NextIndex shr 4] + ', ' + IntToStr(pBuf^ - $8000))
					else
						RawScanStr := RawScanStr + (',' + IntToStr(pBuf^ - $8000))
					;
			end;
			if ThisAIOversample >= DiscardOversamples then begin
				case SmoothingMode of
					smAverage: Inc(IncompleteChannel, pBuf^);
					smMedian, smAverageTheMiddle: Inc(Hist[pBuf^]);
					smAsymmetricStandardDeviations: begin
						Inc(IncompleteChannel, pBuf^);
						Inc(Hist[pBuf^]);
					end;
					smInertia: InertialChData[ThisAICh] := (InertialChData[ThisAICh] * (SmoothingInertia-1) + (SmoothingInertia div 2)) div SmoothingInertia + pBuf^;
				end;
				if pBuf^ < nData then nData := pBuf^;
				if pBuf^ > xData then xData := pBuf^;

				W := Abs(SmallInt(pBuf^ xor $8000));
				if OuterDataOriginal[ThisAICh] < W then OuterDataOriginal[ThisAICh] := W;
			end;
			Inc(pBuf);

			Inc(ThisAIOversample);
			if ThisAIOversample < SamplesPerChannel then Continue;
			ThisAIOversample := 0;

			case SmoothingMode of
				smAverage: begin
					ThisSample := AvgFactor * IncompleteChannel - $8000;
				end;
				smMedian: begin
					ThisSample := 0; //For the compiler.
					W := (SamplesPerChannel - DiscardOversamples) div 2; //Half the data. Once we've gone this far into the histogram, we're at the median.
					for I := nData to xData do begin
						Dec(W, Hist[I]);
						if W < 0 then begin
							ThisSample := I - $8000;
							Break;
						end;
					end;
				end;
				smAverageTheMiddle: begin
					//This will produce a range n2..x2 with the middle half of the data.
					n2 := nData;
					x2 := xData;
					//First, move n2 up from nData to the low mark(25% by default). This is probably somewhere inside the bin, so adjust that bin to leave n2 right at the mark.
					W := iSmoothingCull;
					repeat
						Dec(W, Hist[n2]);
						if W < 0 then Break;
						Inc(n2);
					until False;
					Hist[n2] := -W;
					//Then similarly, move x2 down from xData to the high mark(75% by default), adjusting that bin.
					W := iSmoothingCull;
					repeat
						Dec(W, Hist[x2]);
						if W < 0 then Break;
						Dec(x2);
					until False;
					Hist[x2] := -W;
					//Finally, average the samples n2..x2 together.
					IncompleteChannel := 0;
					W := 0;
					for I := n2 to x2 do begin
						Inc(IncompleteChannel, Hist[I] * I);
						Inc(W, Hist[I]);
					end;
					ThisSample := 1/W * IncompleteChannel - $8000;
				end;
				smAsymmetricStandardDeviations: begin
					//Take the average as a first approximation.
					ThisSample := AvgFactor * IncompleteChannel;
					if nData <> xData then ThisSample := BalanceAsymmetricStandardDeviations(ThisSample, Dev.SmoothingASDReps);
					ThisSample := ThisSample - $8000;
				end;
				smInertia: begin
					ThisSample := InertialChData[ThisAICh] / SmoothingInertia - $8000;
				end;
				else ThisSample := 0;
			end;

			if bLogging and bLogCallbackMinMax then begin
				if ThisAICh = 0 then
					AccScanStr := 'Min/Max: '
				else if ThisAICh = 8 then
					AccScanStr := AccScanStr + #9#9
				else
					AccScanStr := AccScanStr + #9
				;
				AccScanStr := AccScanStr + IntToHex(nData, 4) + '..' + IntToHex(xData, 4);
			end;

			if nData = $0000 then begin
				if xData = $FFFF then begin
					Saturate := stNearest;
					bSaturatedPulse := True;
				end
				else begin
					Saturate := stLower;
					bSaturatedPulse := True;
				end;
			end
			else begin
				if xData = $FFFF then begin
					Saturate := stUpper;
					bSaturatedPulse := True;
				end
				else begin
					Saturate := stNone;
				end;
			end;

			case SmoothingMode of
				smAverage: IncompleteChannel := 0;
				smMedian, smAverageTheMiddle: for I := nData to xData do Hist[I] := 0;
				smAsymmetricStandardDeviations: begin
					IncompleteChannel := 0;
					for I := nData to xData do Hist[I] := 0;
				end;
				smInertia: ;
			end;
			nData := $FFFF;
			xData := $0000;

			iScanSample := ThisAICh;
			if bLogging and bLogCallbackDecimalRaw then RawScanStr := RawScanStr + (',' + FloatToStr(ThisSample));
			with OutChCal[iScanSample] do ThisSample16 := InS16(M * ThisSample + B);
			if bLogging and bLogCallbackDecimalRaw then RawScanStr := RawScanStr + (',' + IntToStr(ThisSample16));

			case Saturate of
				stNone: ;
				stLower: ThisSample16 := Low(ThisSample16);
				stUpper: ThisSample16 := High(ThisSample16);
				stNearest: if ThisSample16 < 0 then
					ThisSample16 := Low(ThisSample16)
				else
					ThisSample16 := High(ThisSample16)
				;
			end;

			IncompleteScan[iScanSample] := ThisSample16;

			//`@ if bRawBypass then IncompleteScan[iScanSample] := LastChData[iScanSample];

			Inc(ThisAICh);
			if ThisAICh < NumChannels then Continue;
			ThisAICh := 0;

			//DAC tick was here.

			if bLogging then begin
				if bLogCallbackMinMax then begin
					ThisNXScanStr := AccScanStr;
					AccScanStr := '';
				end;
				if bLogCallbackDecimalRaw then begin
					ThisRawScanStr := 'Raw Data:' + RawScanStr;
					RawScanStr := '';
				end;
			end;

			if bRejectingData then Continue;

			if bLogging then begin
				if bLogCallbackMinMax then ThreadLogCall(ThisNXScanStr);
				if bLogCallbackDecimalRaw then ThreadLogCall(ThisRawScanStr);
			end;

			PostAIScan(@Dev);
		end;
	end;
end;

begin
	IsMultiThread := True;
end.
