unit LoggingUtils;

interface

uses
  Windows;

//function GetCSIDLPath(CSIDL: Integer): String;

procedure InitializeLogging(bDateTime: Boolean);
procedure FinalizeLogging;
procedure InitializeFDLDACCat;
procedure FinalizeFDLDACCat;
procedure DoLogGains;

function TToStr(T: LongWord): AnsiString;
function LocalTime: AnsiString;
procedure LogCall(NomStr: AnsiString);
procedure ThreadLogCall(NomStr: AnsiString);
procedure LogThreadLogCalls;

procedure SetupFromDACData(pSource: Pointer; Count: Integer; bLogStats: Boolean);
procedure LogRawData(NumSamples: Integer; pData: PWord);
procedure LogDecimalRawData(NumSamples: Integer; pData: PWord);
procedure LogParsedData(NumSamples: Integer; pData: PSmallInt; NumChannels: Integer);

var
  MicrosPerTick: Double;
  GainFactor: array of Integer;
  LastDACCat: array of -1..+2;
  LastDACSection: array of AnsiString;





implementation

uses
  Math, SysUtils, ShlObj, DFileStreams, TheseUtils;

var
  BaseTick, LastLog: Int64;
  LogFil: TDFileStream = nil;
  ThreadLog: array of AnsiString;
  hThreadLogMutex: THandle;
  DACLVal, DACMVal, DACHVal: SmallInt;





{
function GetCSIDLPath(CSIDL: Integer): String;
var
  I: Integer;

begin
  Result := StringOfChar(#0, MAX_PATH + 1);
  if SHGetSpecialFolderPath(Application.Handle, @Result[1], CSIDL, False) then begin
    for I := 1 to Length(Result) do
      if Result[I] = #0 then begin
        SetLength(Result, I - 1);
        Break;
      end
    ;
    Result := IncludeTrailingBackslash(Result);
  end
  else
    Result := ''
  ;
end;
}



procedure InitializeLogging(bDateTime: Boolean);
var
  TimeStr, LogFN: AnsiString;

begin
  hThreadLogMutex := CreateMutex(nil, False, nil);
  QueryPerformanceCounter(BaseTick);
  LastLog := BaseTick;
  FreeAndNil(LogFil);
  TimeStr := LocalTime;
  SetString(LogFN, nil, MAX_PATH);
  SetLength(LogFN, GetTempPath(Length(LogFN), PChar(LogFN)));
  LogFN := IncludeTrailingBackslash(Trim(LogFN));
  if bDateTime then
    LogFN := LogFN + 'USBAIODI ' + StringReplace(StringReplace(TimeStr, '::', ' ', [ rfReplaceAll ]), ':', '', [ rfReplaceAll ]) + '.log'
  else
    LogFN := LogFN + 'USBAIODI.log'
  ;
  try
    LogFil := TDFileStream.Create(LogFN, dfCreateOrReplace + dfReadWrite);
  except
    on E: Exception do AnsiMessageBox(0, E.Message, E.ClassName, MB_ICONEXCLAMATION);
  end;
  LogCall('Log file opened at ' + TimeStr + '.');
end;

procedure FinalizeLogging;
begin
  LogCall('Log file closed at ' + LocalTime + '.');
  FreeAndNil(LogFil);
  CloseHandle(hThreadLogMutex);
end;

procedure InitializeFDLDACCat;
begin

end;

procedure FinalizeFDLDACCat;
var
  I: Integer;

begin
  for I := 0 to High(LastDACCat) do
    LastDACCat[I] := 0
  ;
  for I := 0 to High(LastDACSection) do
    LastDACSection[I] := 'Z'
  ;
end;

procedure DoLogGains;
var
  GainBuf: AnsiString;
  I: Integer;
  
begin
  GainBuf := 'Gains:';
  for I := 0 to High(GainFactor) do
    GainBuf := GainBuf + (' ' + IntToHex(GainFactor[I], 1))
  ;
  LogCall(GainBuf);
end;



function TToStr(T: LongWord): AnsiString;
const
  IntToDigit: array[0..9] of AnsiChar = ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9');

var
  ThousandCycle: Integer;

begin
  ThousandCycle := 0;
  Result := '';
  repeat
    Result := IntToDigit[T mod 10] + Result;
    T := T div 10;
    if T = 0 then Break;

    Inc(ThousandCycle);
    if ThousandCycle >= 3 then begin
      ThousandCycle := 0;
      Result := ' ' + Result;
    end;
  until False;
end;

function GetThreadLogCalls: AnsiString;
var
  I, L: Integer;

begin
  Result := '';
  WaitForSingleObject(hThreadLogMutex, INFINITE);
    L := Length(ThreadLog);
    if L <> 0 then begin
      for I := L-1 downto 0 do begin
        Result := ThreadLog[I] + Result;
      end;
      SetLength(ThreadLog, 0);
    end;
  ReleaseMutex(hThreadLogMutex);
end;

procedure LogCall(NomStr: AnsiString);
var
  Naui: Int64;
  T, dT: Integer;

begin
  if LogFil = nil then Exit;

  QueryPerformanceCounter(Naui);
  T := Round(MicrosPerTick * (Naui-BaseTick));
  dT := Round(MicrosPerTick * (Naui-LastLog));
  NomStr := '[' + TToStr(dT) + #9 + TToStr(T) + '] ' + NomStr + #13#10;
  LastLog := Naui;

  NomStr := GetThreadLogCalls + NomStr;

  LogFil.Write(NomStr[1], Length(NomStr));
end;

procedure LogThreadLogCalls;
var
  BufStr: AnsiString;

begin
  BufStr := GetThreadLogCalls;
  if BufStr = '' then Exit;
  LogFil.Write(BufStr[1], Length(BufStr));
end;

procedure ThreadLogCall(NomStr: AnsiString);
var
  Naui: Int64;
  T, I: Integer;

begin
  QueryPerformanceCounter(Naui);
  T := Round(MicrosPerTick * (Naui-BaseTick));
  NomStr := '[OThread'#9 + TToStr(T) + '] ' + NomStr + #13#10;

  WaitForSingleObject(hThreadLogMutex, INFINITE);
    I := Length(ThreadLog);
    SetLength(ThreadLog, I + 1);
    ThreadLog[I] := NomStr;
  ReleaseMutex(hThreadLogMutex);
end;

function LocalTime: AnsiString;
var
  ST: TSystemTime;

begin
  GetLocalTime(ST);
  with ST do
    Result := Format('%d:%.2d:%.2d::%.2d:%.2d:%.2d', [ wYear, wMonth, wDay, wHour, wMinute, wSecond ])
  ;
end;



//Final data logging functions.
procedure SetupFromDACData(pSource: Pointer; Count: Integer; bLogStats: Boolean);
type
  TDADRelation = record
    DACVal: SmallInt;
    Wt: Integer;
  end;

var
  iSample, iRel, H, I, L: Integer;
  LogStr: AnsiString;
  LastDACData: array of SmallInt;
  DADRel: array of TDADRelation;
  LastLastDACSection: AnsiString;

  function FindDADRelI(ByDACVal: SmallInt): Integer;
  var
    I: Integer;

  begin
    Result := -1;
    for I := 0 to High(DADRel) do
      if DADRel[I].DACVal = ByDACVal then begin
        Result := I;
        Break;
      end
    ;
  end;

  procedure SwapDADRels(iA, iB: Integer);
  var
    DADRelBuf: TDADRelation;

  begin
    DADRelBuf := DADRel[iA];
    DADRel[iA] := DADRel[iB];
    DADRel[iB] := DADRelBuf;
  end;

  function ListDADRels: AnsiString;
  var
    I: Integer;

  begin
    Result := '[' + IntToHex(Word(DADRel[0].DACVal), 4);
    for I := 1 to High(DADRel) do
      Result := Result + (', ' + IntToHex(Word(DADRel[I].DACVal), 4))
    ;
    Result := Result + ']';
  end;

begin
  //Copy the DAC data for ourselves.
  SetLength(LastDACData, Count);
  Move(pSource^, LastDACData[0], SizeOf(LastDACData[0])*Count);

  //Histogram the DAC data, so we know when it should be zero, high, low.
  SetLength(DADRel, 0);
  for iSample := 0 to Count-1 do begin
    iRel := FindDADRelI(LastDACData[iSample]);
    if iRel >= 0 then begin
      Inc(DADRel[iRel].Wt);
    end
    else begin
      iRel := Length(DADRel);
      SetLength(DADRel, iRel + 1);
      DADRel[iRel].DACVal := LastDACData[iSample];
      DADRel[iRel].Wt := 1;
    end;
  end;

  //Stuff the zero, high, low values based on the histogram.
  L := Length(DADRel);
  case L of
    0: begin
      LogCall('No unique values in []. Something went wrong.');
      DACLVal := 0;
      DACMVal := 0;
      DACHVal := 0;
    end;
    1: begin
      LogCall('Only 1 unique value in ' + ListDADRels + '. Slopes won''t be calculable.');
      if DADRel[0].DACVal > DADRel[1].DACVal then SwapDADRels(0, 1);
      DACMVal := DADRel[0].DACVal;
      DACLVal := DACMVal - 1;
      DACHVal := DACMVal + 1;
    end;
    2: begin
      LogCall('Only 2 unique values in ' + ListDADRels + '. Slopes are supported, barely.');
      if DADRel[0].DACVal > DADRel[1].DACVal then SwapDADRels(0, 1);
      DACLVal := DADRel[0].DACVal;
      DACHVal := DADRel[1].DACVal;
      DACMVal := (DACLVal + DACHVal) div 2;
    end;
    else begin
      if L > 3 then begin
        LogStr := IntToHex(L) + ' unique values in ' + ListDADRels + ', truncating to ';
        H := L-1;
        repeat
          for I := 0 to H-1 do
            if DADRel[I].Wt < DADRel[I + 1].Wt then
              SwapDADRels(I, I + 1)
          ;
          Dec(H);
        until H < 3;
        SetLength(DADRel, H+1);
        LogStr := LogStr + ListDADRels + '.';
        LogCall(LogStr);
      end
      else begin
        LogCall('3 unique values in ' + ListDADRels + '.');
      end;
      if DADRel[0].DACVal > DADRel[1].DACVal then SwapDADRels(0, 1);
      if DADRel[1].DACVal > DADRel[2].DACVal then SwapDADRels(1, 2);
      if DADRel[0].DACVal > DADRel[1].DACVal then SwapDADRels(0, 1);
      DACLVal := DADRel[0].DACVal;
      DACMVal := DADRel[1].DACVal;
      DACHVal := DADRel[2].DACVal;
    end;
  end;
  SetLength(DADRel, 0);

  LogCall('DAC targets are ' + IntToHex(Word(DACLVal), 4) + ', ' + IntToHex(Word(DACMVal), 4) + ', ' + IntToHex(Word(DACHVal), 4) + '.');

  //Categorize the DAC values.
  SetLength(LastDACCat, Count);
  SetLength(LastDACSection, Count+1);
  LastDACSection[0] := 'Z';
  LastLastDACSection := '';
  for iSample := 0 to Count-1 do begin
    if LastDACData[iSample] = DACMVal then begin
      LastDACCat[iSample] := 0;
      if (LastLastDACSection = '') or (LastLastDACSection = 'B1') then
        LastDACSection[iSample+1] := 'B1'
      else if (LastLastDACSection = 'B2') or (LastLastDACSection = 'C') then
        LastDACSection[iSample+1] := 'C'
      else if (LastLastDACSection = 'B3') or (LastLastDACSection = 'H') then
        LastDACSection[iSample+1] := 'H'
      else
        LastDACSection[iSample+1] := 'Z'
      ;
    end
    else if LastDACData[iSample] = DACLVal then begin
      LastDACCat[iSample] := -1;
      LastDACSection[iSample+1] := 'B3';
    end
    else if LastDACData[iSample] = DACHVal then begin
      LastDACCat[iSample] := +1;
      LastDACSection[iSample+1] := 'B2';
    end
    else begin
      LastDACCat[iSample] := 2;
      LastDACSection[iSample+1] := '?';
    end;
    LastLastDACSection := LastDACSection[iSample+1];
  end;
  SetLength(LastDACData, 0);

  if bLogStats then begin
    LogStr := IntToStr(LastDACCat[0]);
    for I := 1 to High(LastDACCat) do
      LogStr := LogStr + (', ' + IntToStr(LastDACCat[I]))
    ;
    LogCall('DAC categorizations are [' + LogStr + '].');
  end;
end;

procedure LogRawData(NumSamples: Integer; pData: PWord);
var
  iReadback: Integer;
  LogStr: AnsiString;

begin
  LogStr := 'Data(' + IntToHex(NumSamples) + ' samples):';
  for iReadback := 0 to NumSamples-1 do begin
    if (iReadback and $F) = 0 then
      LogStr := LogStr + #13#10
    else
      LogStr := LogStr + #9
    ;
    LogStr := LogStr + IntToHex(pData^, 4);
    Inc(pData);
  end;
  LogCall(LogStr);
end;

procedure LogDecimalRawData(NumSamples: Integer; pData: PWord);
var
  iReadback: Integer;
  LogStr: AnsiString;

begin
  LogStr := 'Data(' + IntToHex(NumSamples) + ' samples):';
  for iReadback := 0 to NumSamples-1 do begin
    if (iReadback and $F) = 0 then
      LogStr := LogStr + #13#10
    else
      LogStr := LogStr + #9
    ;
    LogStr := LogStr + IntToStr(PSmallInt(pData)^);
    Inc(pData);
  end;
  LogCall(LogStr);
end;

procedure LogParsedData(NumSamples: Integer; pData: PSmallInt; NumChannels: Integer);
var
  NumScans, iRel, iTissue: Integer;
  LogStr: AnsiString;
  DADRel: array[-1..+1] of record
    tData: Integer;
    Wt: Integer;
  end;
  LAvg, ZAvg, HAvg, DAvg, ZAAvg, DAAvg: Integer;

  function AvgFromDADRel(iDADRel: Integer): Integer;
  begin
    with DADRel[iDADRel] do
      if Wt = 0 then
        Result := 0
      else
        Result := Round(tData / Wt)
    ;
  end;

  procedure CollectRelData(iADCh: Integer);
  var
    iScan: Integer;
    pThisData: PSmallInt;

  begin
    with DADRel[-1] do begin tData := 0; Wt := 0; end;
    with DADRel[ 0] do begin tData := 0; Wt := 0; end;
    with DADRel[+1] do begin tData := 0; Wt := 0; end;
    pThisData := pData;
    Inc(pThisData, iADCh + NumChannels);
    for iScan := 1 to NumScans-1 do begin
      iRel := LastDACCat[Max(0, iScan-1)];
      if iRel < 2 then
        with DADRel[iRel] do begin
          tData := tData + pThisData^;
          Inc(Wt);
        end
      ;
      Inc(pThisData, NumChannels);
    end;
    LAvg := AvgFromDADRel(-1); //with DADRel[-1] do if Wt = 0 then LAvg := 0 else LAvg := Round(tData / Wt);
    ZAvg := AvgFromDADRel( 0); //with DADRel[ 0] do if Wt = 0 then ZAvg := 0 else ZAvg := Round(tData / Wt);
    HAvg := AvgFromDADRel(+1); //with DADRel[+1] do if Wt = 0 then HAvg := 0 else HAvg := Round(tData / Wt);
  end;

begin
  try
    if Length(GainFactor) < NumChannels then Exit;

    NumScans := Min(NumSamples div NumChannels, Length(LastDACCat));

    LogStr := 'Data(' + IntToHex(NumScans) + ' scans/DACs):';
    for iTissue := 0 to 5 do begin
      LogStr := LogStr + (#13#10'[' + IntToHex(iTissue) + ']: ');
      CollectRelData(iTissue * 2);
      LogStr := LogStr + ('Ch' + IntToHex(iTissue  ) + ': ' + ModifHex(LAvg) + '..' + ModifHex(ZAvg) + '..' + ModifHex(HAvg) + ', delta=' + ModifHex(HAvg - LAvg) + ', ');
      CollectRelData(iTissue * 2 + 1);
      LogStr := LogStr + ('Ch' + IntToHex(iTissue+8) + ': ' + ModifHex(LAvg) + '..' + ModifHex(ZAvg) + '..' + ModifHex(HAvg) + ', delta=' + ModifHex(HAvg - LAvg));
    end;
    LogCall(LogStr);

    LogStr := 'Parsed(decimal):'#13#10'Tissue[],mV,dmV,A,dA,Ohms';
    for iTissue := 0 to 5 do begin
      CollectRelData(iTissue * 2 + 1);
      ZAAvg := ZAvg;
      DAAvg := Abs(HAvg - LAvg);
      CollectRelData(iTissue * 2);
      DAvg := Abs(HAvg - LAvg);
      LogStr := LogStr + (#13#10 + IntToStr(iTissue)
        + ',' + Modif     (20000/$10000 / GainFactor[iTissue*2  ] * ZAvg )
        + ',' + FloatToStr(20000/$10000 / GainFactor[iTissue*2  ] * DAvg )
        + ',' + Modif     (20000/$10000 / GainFactor[iTissue*2+1] * ZAAvg)
        + ',' + FloatToStr(20000/$10000 / GainFactor[iTissue*2+1] * DAAvg)
      );
      if DAAvg = 0 then
        LogStr := LogStr + ',"Undefined"'
      else
        LogStr := LogStr + (',' + FloatToStr(1000 * DAvg / DAAvg))
      ;
    end;
    LogCall(LogStr);
  except
    on E: Exception do LogCall('Exception in LogParsedData: ' + E.ClassName + ' = ' + E.Message);
  end;
end;

end.
